/**
 * @copyright Copyright (c) Open-Xchange GmbH, Germany <info@open-xchange.com>
 * @license AGPL-3.0
 *
 * This code is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with OX App Suite. If not, see <https://www.gnu.org/licenses/agpl-3.0.txt>.
 *
 * Any use of the work other than as authorized under this license or copyright law is prohibited.
 */

const { platform } = require('node:process')
const { expect } = require('@playwright/test')
const util = require('./util')
const codecept = require('codeceptjs')

const baseURL = util.getURLRoot()

module.exports = function () {
  return actor({

    amOnLoginPage () {
      this.amOnPage(baseURL)
    },

    openApp (appName) {
      this.click('~Navigate to:')
      this.click(appName, '.apps')
    },

    /**
     * This simplifies the input of mail addresses into input fields.
     * There must be an array of users defined in the AppSuite helper configuration.
     * If you want to fill in a mailaddress of such a user you can simply use this function.
     *
     * @param selector {string} the selector of an editable field
     * @param userIndex {number} the users position in the users array provided via helper config
     */
    insertMailaddress (selector, userIndex) {
      const config = codecept.config.get()
      const users = config.helpers.AppSuite.users
      if (userIndex < 0 || userIndex >= users.length) userIndex = 0
      const user = users[userIndex]
      this.fillField(selector, user.mail)
    },

    /**
     * Logs in the user with the specified URL parameters and options.
     * If the URL parameters are an object, it will be converted to an empty array.
     * The options parameter is an object that can contain a 'user' property.
     * If the 'user' property is not provided, the first user from the 'users' container will be used.
     * The login process includes making an API request to the login endpoint with the user's credentials.
     * After successful login, the page will be navigated to the specified URL with the URL parameters.
     * If the 'isDeepLink' option is true, the URL will be constructed with a '#' prefix.
     * The login process also includes waiting for the page to load and checking for the presence of the app.
     * @param {string[]|object} urlParams - The URL parameters as an array or object.
     * @param {object} options - The login options.
     * @returns {Promise<void>} - A promise that resolves when the login process is complete.
     */
    async login (urlParams, options) {
      if (urlParams && !urlParams.length) {
        // looks like an object, not string or array
        options = urlParams
        urlParams = []
      }
      urlParams = [].concat(urlParams || [])
      options = Object.assign({ wait: true }, options)
      let user = options.user || require('codeceptjs').container.support('users')[0]

      if (typeof user === 'undefined') user = {}
      if (user.toJSON) user = user.toJSON()
      const loginName = process.env.PROVISIONING_API === 'reseller' ? user.name : `${user.name}${user.context.id ? '@' + user.context.id : ''}`
      await this.makeApiRequest('POST', `${baseURL}/api/login`, {
        params: { action: 'login', client: 'open-xchange-appsuite' },
        form: {
          name: loginName,
          password: user.password
        }
      })
      this.wait(0.5)
      this.amOnPage(`${baseURL}/` + (options.isDeepLink ? '#' : '#!!') + (urlParams.length === 0 ? '' : '&' + urlParams.join('&')))
      this.waitForInvisible('#background-loader.busy', util.getLoginTimeout())
      if (options.wait) await this.waitForApp()
    },

    async logout () {
      await this.makeApiRequest('POST', `${baseURL}/api/login`, { params: { action: 'logout' } })
      this.amOnPage('about:blank')
    },

    waitForNetworkTraffic () {
      this.waitForDetached({ css: '[aria-label="Currently refreshing"]' }, 15)
    },

    // Wait for focus in playwright
    /**
     * Waits for the specified element to receive focus.
     *
     * @param {string} locator - The locator of the element. Only accepts css selectors.
     */
    waitForFocus (locator) {
      this.waitForElement(locator)
      this.waitForVisible(locator)
      return this.usePlaywrightTo('wait for focus', async ({ page }) => {
        await expect(page.locator(locator)).toBeFocused()
        this.wait(0.5)
      })
    },

    triggerRefresh () {
      this.waitForVisible('~Refresh')
      this.click('~Refresh')
      this.waitForInvisible('~Currently refreshing', 10)
      this.waitForVisible('~Refresh')
    },

    grabBackgroundImageFrom (selector) {
      return this.grabCssPropertyFrom(selector, 'background-image').then(backgroundImage => {
        backgroundImage = Array.isArray(backgroundImage) ? backgroundImage[0] : backgroundImage
        return backgroundImage || 'none'
      })
    },

    clickDropdown (text) {
      this.waitForVisible('.dropdown.open .dropdown-menu, .primary-action > .open .dropdown-menu')
      this.waitForText(text, 10, '.dropdown.open .dropdown-menu, .primary-action > .open .dropdown-menu')
      return this.click(text, '.dropdown.open .dropdown-menu, .primary-action > .open .dropdown-menu')
    },

    clickToolbar (selector, timeout = 5) {
      this.waitForElement(`.classic-toolbar [aria-label^="${selector}"]`, timeout)
      this.click(`~${selector}`, '.classic-toolbar')
    },

    clickPrimary (text) {
      const button = locate('.primary-action .btn-primary').withText(text).as(`Primary Button: ${text}`)
      this.waitForVisible(button)
      this.click(button)
    },

    openFolderMenu (folderName) {
      const item = `.folder-tree [title*="${folderName}"]`
      this.waitForVisible(item)
      this.rightClick(item)
    },

    changeTheme ({ theme }) {
      // "White" "Blue" "Steel gray" "Dark" "Mountains" "Beach" "City" "Blue Sunset"
      this.click('[aria-label="Settings"]')
      this.waitForVisible('#topbar-settings-dropdown')
      this.click(`div[aria-label="Themes"] a[title="${theme}"]`)
      this.waitForNetworkTraffic()
      this.pressKey('Escape')
      this.wait(0.5)
    },

    /**
     * Selects the text in a text field or contenteditable element from cursor position to beginning of line.
     */
    selectLine () {
      const shortcut = platform === 'darwin' ? ['Shift', 'CommandOrControl', 'ArrowLeft'] : ['Shift', 'Home']
      this.pressKey(shortcut)
    }
  })
}
