import { describe, it, expect } from '@jest/globals'
import * as rollup from 'rollup'
import gettextPlugin from '../dist/index'

describe('compile po file', function () {
  async function render (opts = {}) {
    const bundle = await rollup.rollup({
      input: {
        'test/mySimpleModule': 'gettext'
      },
      plugins: [gettextPlugin(Object.assign({
        poFiles: 'spec/fixtures/i18n/*.po',
        defaultDictionary: 'test/mySimpleModule',
        defaultLanguage: 'de_DE'
      }, opts))]
    })
    return await bundle.generate({ })
  }
  function getTranslationsFrom (mod) {
    return JSON.parse(mod.code.match(/ = ({\n.*headers:\s*headers,[^}]+})/m)[1].replace(/headers:\s*headers,/, ''))
  }
  describe('using a simple example', function () {
    it('should generate a loadable ES module', async function () {
      const { output } = await render()
      const simpleModule = output.find(chunk => /SimpleModule/.test(chunk.fileName))
      const alternativeModule = output.find(chunk => /alternative/.test(chunk.fileName))
      expect(simpleModule.code.match(/Language:\s*"([^"]+)/)[1]).toEqual('de_DE')
      expect(alternativeModule.code.match(/Language:\s*"([^"]+)/)[1]).toEqual('de_DE')
    })
    it('missing translation should be filtered from dictionary', async function () {
      const { output } = await render()

      expect(output.map(c => c.code).join('\n').indexOf('part of computer\\u0000Drive contains %1$d item\\u0001Drive contains %1$d items')).toBeLessThan(0)
    })
    describe('items with no context', function () {
      it('should render keys with singular form only', async function () {
        const { output } = await render()
        const simpleModule = output.find(chunk => /SimpleModule/.test(chunk.fileName))
        const translation = getTranslationsFrom(simpleModule)
        expect(translation['\x00translate me!\x01'][0]).toEqual('übersetze mich!')
      })
      it('should render keys with singular and plural forms', async function () {
        const { output } = await render()
        const simpleModule = output.find(chunk => /SimpleModule/.test(chunk.fileName))
        const translation = getTranslationsFrom(simpleModule)
        expect(translation['\x00Box contains %1$d item\x01Box contains %1$d items'][0]).toEqual('Box enthält %1$d Elemente')
        expect(translation['\x00Box contains %1$d item\x01Box contains %1$d items'][1]).toEqual('Box enthält %1$d Elemente')
      })
      it('should render keys with singualar and plural forms with only one field specified in po file', async function () {
        // this might happen with japanese translations, po file initially contains multiple msg_str entries,
        // but those are removed by the translators
        const { output } = await render()
        const simpleModule = output.find(chunk => /SimpleModule/.test(chunk.fileName))
        const translation = getTranslationsFrom(simpleModule)
        const result = translation['\x00Partially translated plural string\x01Partially translated plural strings']
        expect(result).toBeInstanceOf(Array)
        expect(result.length).toEqual(1)
        expect(result[0]).toEqual('Teilweise übersetzte Pluralzeichenkette')
      })
    })
    describe('items with context', function () {
      it('should render keys with singular form only', async function () {
        const { output } = await render()
        const simpleModule = output.find(chunk => /SimpleModule/.test(chunk.fileName))
        const translation = getTranslationsFrom(simpleModule)
        expect(translation['describe state\x00Empty folder\x01'][0]).toEqual('Leerer Ordner')
        expect(translation['perform action\x00Empty folder\x01'][0]).toEqual('Leere Ordner')
      })
      it('should render keys with singular and plural forms', async function () {
        const { output } = await render()
        const simpleModule = output.find(chunk => /SimpleModule/.test(chunk.fileName))
        const translation = getTranslationsFrom(simpleModule)
        expect(translation['\x00Box contains %1$d item\u0001Box contains %1$d items'][0]).toEqual('Box enthält %1$d Elemente')
        expect(translation['\x00Box contains %1$d item\u0001Box contains %1$d items'][1]).toEqual('Box enthält %1$d Elemente')
      })
    })
    describe('messages with fuzzy flag', function () {
      it('should be filtered from catalog by default', async function () {
        const { output } = await render()
        const simpleModule = output.find(chunk => /SimpleModule/.test(chunk.fileName))
        const translation = getTranslationsFrom(simpleModule)
        expect(translation['\x00A message with fuzzy translation\x01']).toBeUndefined()
      })

      it('should be included in catalog if includeFuzzy option is set', async function () {
        const { output } = await render({ includeFuzzy: true })
        const simpleModule = output.find(chunk => /SimpleModule/.test(chunk.fileName))
        const translation = getTranslationsFrom(simpleModule)
        expect(translation['\x00A message with fuzzy translation\x01']).toBeInstanceOf(Array)
      })

      it('should dynamically load translation modules', async () => {
        const bundle = await rollup.rollup({
          input: 'test/mySimpleModule',
          plugins: [gettextPlugin({
            poFiles: 'spec/fixtures/i18n/*.po',
            defaultDictionary: 'test/mySimpleModule',
            defaultLanguage: 'de_DE'
          })]
        })
        expect((await bundle.generate({})).output.map(c => c.fileName)).toHaveLength(4)
      })
    })
  })
  describe('with mesages pointing to different modules', function () {
    it('should put message into each referenced module', async function () {
      const { output } = await render()
      const simpleModule = output.find(chunk => /SimpleModule/.test(chunk.fileName))
      const translationSimpleModule = getTranslationsFrom(simpleModule)
      const alternativeModule = output.find(chunk => /alternative/.test(chunk.fileName))
      const translationAlternativeModule = getTranslationsFrom(alternativeModule)
      expect(translationSimpleModule['\x00String with " in it\x01']).toBeInstanceOf(Array)
      expect(translationAlternativeModule['\x00String with " in it\x01']).toBeInstanceOf(Array)
    })
  })
  it('should not extract obsolete messages', async function () {
    const { output } = await render()
    const alternativeModule = output.find(chunk => /alternative/.test(chunk.fileName))
    const translation = getTranslationsFrom(alternativeModule)
    expect(translation['\x00unused message\x01']).not.toBeInstanceOf(Array)
  })
})
