import { describe, it, expect, afterEach } from '@jest/globals'
import { build, createServer } from 'vite'
import vitePluginOxExternals from '../../src/index'
import { getFileFromBundle, getPort, getDirname } from '../util'

const PORT = getPort()
const __dirname = getDirname(import.meta.url)

describe('Dynamic imports', function () {
  let server

  afterEach(async function () {
    if (server) {
      await server.close()
      await server.httpServer.close()
      server = null
    }
  })

  it('works', async function () {
    const bundle = await build({
      root: __dirname,
      logLevel: 'silent',
      build: {
        write: false,
        rollupOptions: {
          output: {
            entryFileNames: 'entries/[name].js'
          }
        }
      },
      plugins: [vitePluginOxExternals({
        prefix: '$'
      })]
    })

    const indexJS = getFileFromBundle('index.js', bundle)
    expect(indexJS.code).toContain('import("../module/external.js")')
    expect(indexJS.dynamicImports).toEqual(['module/external.js'])
  })

  it('works with minified disable', async function () {
    const bundle = await build({
      root: __dirname,
      logLevel: 'silent',
      build: {
        minify: false,
        write: false,
        rollupOptions: {
          output: {
            entryFileNames: 'entries/[name].js'
          }
        }
      },
      plugins: [vitePluginOxExternals({
        prefix: '$'
      })]
    })

    const indexJS = getFileFromBundle('index.js', bundle)
    expect(indexJS.code).toContain('import("../module/external.js")')
    expect(indexJS.dynamicImports).toEqual(['module/external.js'])
  })

  it('skips module names in strings', async function () {
    const bundle = await build({
      root: __dirname,
      logLevel: 'silent',
      build: {
        minify: false,
        write: false,
        rollupOptions: {
          output: {
            entryFileNames: 'entries/[name].js'
          }
        }
      },
      plugins: [vitePluginOxExternals({
        prefix: '$'
      })]
    })

    const indexJS = getFileFromBundle('index.js', bundle)
    expect(indexJS.code).toContain('Not replace: import("module/external.js")')
    expect(indexJS.dynamicImports).toEqual(['module/external.js'])
  })

  it('works in dev mode', async function () {
    server = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginOxExternals({
        prefix: '$'
      })]
    })
    await server.listen(PORT)

    const res = await fetch(`http://localhost:${PORT}/index.js`)
    expect(await res.text()).toEqual(expect.stringContaining(`document.addEventListener('resize', async () => {
  const { default: mod } = import("/module/external.js")
  console.log('Not replace: import("module/external.js")', mod)
})
`))

    await server.close()
  })
})
