import { describe, it, expect, afterEach } from '@jest/globals'
import { build, createServer } from 'vite'
import vitePluginOxExternals from '@open-xchange/vite-plugin-ox-externals'

import vitePluginOxManifests from '../../dist/index'
import { getFileFromBundle, getPort, getDirname } from '../util'

const PORT = getPort()
const __dirname = getDirname(import.meta.url)

describe('Multiple manifest scenario', function () {
  let server

  afterEach(async function () {
    if (server) {
      await server.close()
      server = null
    }
  })

  it('works', async function () {
    const bundle = await build({
      root: __dirname,
      logLevel: 'silent',
      build: {
        write: false,
        rollupOptions: {
          input: {}
        }
      },
      plugins: [vitePluginOxExternals({
        prefix: '$'
      }), vitePluginOxManifests({

      })]
    })

    const manifestFile = getFileFromBundle('manifest', bundle)
    const manifests = JSON.parse(manifestFile.source)
    expect(Object.keys(manifests)).toHaveLength(2)

    expect(manifests['entry1.js'].meta.manifests[0].namespace).toEqual('test')

    expect(manifests['sub/entry2.js'].meta.manifests).toHaveLength(3)
    expect(manifests['sub/entry2.js'].meta.manifests.find(m => m.namespace === 'other').namespace).toEqual('other')
    expect(manifests['sub/entry2.js'].meta.manifests.find(m => m.namespace === 'test').namespace).toEqual('test')
    expect(manifests['sub/entry2.js'].meta.manifests.find(m => m.namespace === 'settings').namespace).toEqual('settings')
  })

  it('works in dev mode', async function () {
    server = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginOxManifests({

      }), vitePluginOxExternals({
        prefix: '$'
      })]
    })
    await server.listen(PORT)

    const res = await fetch(`http://localhost:${PORT}/api/manifest.json`)
    const manifests = await res.json()

    expect(manifests).toHaveLength(4)

    const entry1Index = manifests.findIndex(m => m.path.indexOf('entry1') >= 0)
    expect(manifests[entry1Index].namespace).toEqual('test')
    expect(manifests[entry1Index].path).toContain('entry1')

    const otherNSIndex = manifests.findIndex(m => m.namespace === 'other')
    expect(manifests[otherNSIndex].namespace).toEqual('other')
    expect(manifests[otherNSIndex].path).toContain('entry2')

    const lastIndex = manifests.findIndex(({ path, namespace }) => path.indexOf('entry2') >= 0 && namespace === 'test')
    expect(manifests[lastIndex].namespace).toEqual('test')
    expect(manifests[lastIndex].path).toContain('entry2')

    const settingsIndex = manifests.findIndex(m => m.namespace === 'settings')
    expect(manifests[settingsIndex].namespace).toEqual('settings')
    expect(manifests[settingsIndex].path).toContain('entry2')
  })
})
