import type http from 'node:http'
import type http2 from 'node:http2'

import type { ProxyConfig } from '../utils/types.js'
import { getMessage } from '../utils/util.js'
import logger from '../utils/logger.js'

export function createPath (path: string, config: ProxyConfig): string {
  const concatenatedPath = config.prependPath + path
  try {
    return config.rewrite(concatenatedPath)
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
  } catch (err) {
    logger.warn(`Could not rewrite path "${concatenatedPath}". Falling back to original path`)
    return concatenatedPath
  }
}

export function createHeaders (req: http.IncomingMessage, config: ProxyConfig): http.OutgoingHttpHeaders {
  const headers: http.OutgoingHttpHeaders = {}
  // forwarded request falls back to http 1.x, so we must have a valid host header
  if (config.changeOrigin) {
    const { hostname, port } = config.url
    headers.host = port ? `${hostname}:${port}` : hostname
  } else {
    const authority: string = req.headers[':authority'] as string
    headers.host = authority || req.headers.host
  }
  return headers
}

export async function pipe (proxyRes: http.IncomingMessage, res: http.ServerResponse | http2.Http2ServerResponse): Promise<void> {
  return new Promise(resolve => {
    const { statusCode = 0 } = proxyRes
    let bytesWritten = 0

    function setHeaders (): void {
      if (!bytesWritten) {
        res.statusCode = statusCode
        for (const [key, value] of Object.entries(proxyRes.headers)) {
          res.setHeader(key, value!)
        }
      }
    }

    function onDrain (): void {
      proxyRes.resume()
    }

    // pipe the response data from proxyRes to res
    proxyRes
      .on('data', (buf: Buffer) => {
        setHeaders()
        bytesWritten += buf.length
        // TS is not able to infer the `write` method from both response classes (http/http2)
        if (!(res as http.ServerResponse).write(buf)) {
          proxyRes.pause()
        }
      })
      .on('end', () => {
        try {
          setHeaders()
        } catch (err) {
          logger.error(`Could not proxy headers: ${getMessage(err)}`)
        }

        // trailers are only there, when version is compliant
        // proxyRes.trailers contains headers that are sent at the end of the request
        if (proxyRes.trailers) res.addTrailers(proxyRes.trailers)
        res.end()
        resolve()
      })
      .on('close', () => {
        res.off('drain', onDrain)
      })

    res.on('drain', onDrain)
  })
}
